/*
 * main.cxx
 *
 * A simple H.323 MCU
 *
 * Copyright (c) 2000 Equivalence Pty. Ltd.
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is Portable Windows Library.
 *
 * The Initial Developer of the Original Code is Equivalence Pty. Ltd.
 *
 * Portions of this code were written by Post Increment (http://www.postincrement.com) 
 * with the assistance of funding from Citron Networks (http://www.citron.com.tw)
 *
 * Portions are Copyright (C) 1993 Free Software Foundation, Inc.
 * All Rights Reserved.
 *
 * Contributor(s): Derek J Smithies (derek@indranet.co.nz)
 *                 ------------------------------
 *
 * $Log: main.cxx,v $
 * Revision 2.6  2004/04/02 01:10:52  csoutheren
 * Fixed unitialised variable
 *
 * Revision 2.5  2004/03/31 03:40:25  csoutheren
 * Added various codecs back in
 *
 * Revision 2.4  2004/03/31 03:36:38  csoutheren
 * Fixed problem with user indication messages
 * Fixed problems with room listener and unlisten
 *
 * Revision 2.3  2004/03/30 04:02:54  csoutheren
 * Fixed problem with not using incoming alias for default room name
 *
 * Revision 2.2  2004/03/23 11:40:06  csoutheren
 * Fixed problem where deleting map element in-place causes crash at end of call
 * Fixed problem where referencing map by iterator rather than ID
 * Fixed code formatting problems
 *
 * Revision 2.1  2004/03/11 20:49:44  csoutheren
 * Removed warnings
 *
 * Revision 2.0  2004/03/08 02:06:24  csoutheren
 * Totally rewritten to use new connection locking mecahnism
 * Added ability to monitor conferences
 * Added initial support for H.323 MCU messages
 * Thanks to Citron Networks for supporting this work
 *
 * Revision 1.75  2004/02/23 23:18:04  csoutheren
 * Fixed iLBC compile problems
 * Fixed NO_VIDEO compile problems
 *
 * Revision 1.74  2004/01/02 02:53:11  csoutheren
 * Thanks to Michal Zygmuntowicz for these changes
 *    iLBC codec support,
 *    ability to specify a gatekeeper password.
 *
 * Revision 1.73  2003/10/05 21:03:02  dereksmithies
 * The H263 codecs should only be enabled when the "v" option is specified.
 *
 * Revision 1.72  2003/08/27 23:19:43  dereksmithies
 * Fix support for H263 codecs. Thanks to Frankie Chiang for pointing it out.
 *
 * Revision 1.71  2003/08/06 08:31:50  dereksmithies
 * Second attempt to get it right. Improves efficiency & audio handling.
 *
 * Revision 1.70  2003/08/05 22:47:39  dereksmithies
 * Apply code from Martin Andre to give single stream functionality. Thanks!
 *
 * Revision 1.69  2003/07/13 22:01:25  dereksmithies
 * Apply patch from Martin Andre to stop crash when sending large video. Thanks!
 *
 * Revision 1.68  2003/06/17 03:05:24  dereksmithies
 * Add H263 codec.
 *
 * Revision 1.67  2003/06/12 23:20:38  dereksmithies
 * Bug fix from Dave, related to delays in VideoEncoding. Thanks!
 *
 * Revision 1.66  2003/05/01 07:02:48  robertj
 * Added basic string capability for user input
 *
 * Revision 1.65  2003/05/01 06:30:43  robertj
 * Added propagation of user indications from a connection to all other
 *   connections in a room.
 * Minor API clean up.
 *
 * Revision 1.64  2003/03/23 20:58:18  dereks
 * Add no video patch from Ivanov - Thanks.
 *
 * Revision 1.63  2003/03/17 08:23:04  robertj
 * Removed openh323 versions of videoio.h classes as PVideoOutputDevice
 *   descendants for NULL and PPM files added to PWLib.
 *
 * Revision 1.62  2003/02/10 09:29:15  rogerh
 * Check that a room still exists when reading/writing audio and video.
 * This fixes a problem where the video compression thread was reading
 * from a room which had been deleted.
 *
 * Revision 1.61  2003/02/06 22:29:09  robertj
 * Fixed MSVC warning.
 *
 * Revision 1.60  2003/02/06 02:43:49  rogerh
 * There is now a video buffer for each 'room', which stops people in
 * different rooms from seeing each other. Bug pointed out by Damien Sandras.
 * Will need for fix leaks and delete the video buffer when a room empties.
 *
 * Revision 1.59  2002/12/16 08:57:12  robertj
 * Fixed problem with spoken list crashing with video, thanks Bob Lindell
 *
 * Revision 1.58  2002/11/21 07:58:32  robertj
 * Fixed stupid mistake in removing make call function.
 *
 * Revision 1.57  2002/11/21 07:54:52  robertj
 * Removed redundent and incorrect function for making outgoing call.
 *
 * Revision 1.56  2002/11/13 10:23:57  rogerh
 * Enable Speex codec by default, now it is part of OpenH323.
 *
 * Revision 1.55  2002/11/10 08:12:42  robertj
 * Moved constants for "well known" ports to better place (OPAL change).
 *
 * Revision 1.54  2002/10/31 00:55:42  robertj
 * Enhanced jitter buffer system so operates dynamically between minimum and
 *   maximum values. Altered API to assure app writers note the change!
 *
 * Revision 1.53  2002/10/16 10:01:28  rogerh
 * Add Speex 8k and 15k codecs.
 *
 * Revision 1.52  2002/06/14 08:05:08  dereks
 * Added tweak to support operation behind a NAT that forwards ports
 * This will use TranslateTCPAddress to change the address reported
 * to connections that occur from outside the 192. subnet. Thanks Sahai.
 *
 *
 * Revision 1.51  2002/04/18 10:54:34  rogerh
 * Fix bug in audio mixing reportde by Bob Lindell <lindell@isi.edu>
 * AudioBuffer::Read() had been optimised to do reading and mixing and
 * contained memset()'s which were incorrect. Clean up the code with
 * a proper Read() method and an optimised ReadAndMix() method.
 *
 * Revision 1.50  2002/04/06 00:54:43  rogerh
 * Clean up the videoBufferSize code
 *
 * Revision 1.49  2002/04/06 00:15:14  rogerh
 * Fix bugs in the Clear() method.
 *
 * Revision 1.48  2002/02/17 08:46:20  rogerh
 * Add simple wildcard of "*" for --audio-loopback to enable loopback
 * in all rooms
 *
 * Revision 1.47  2002/02/17 08:30:17  rogerh
 * The new openh323 library checks for getLastReadCount
 *
 * Revision 1.46  2002/02/07 12:19:15  rogerh
 * add --listenport option. Requested by Damien
 *
 * Revision 1.45  2002/01/16 10:19:45  rogerh
 * Tidy code and screen alignment
 *
 * Revision 1.44  2001/12/02 08:28:39  rogerh
 * Update help for audio-loopback
 *
 * Revision 1.43  2001/12/02 08:17:57  rogerh
 * Change --audio-loopback to take a room name. Only connections to the
 * specified room will have audio loopback applied. The remainder of the rooms
 * in OpenMCU will work as normal.
 * Submitted by Malcolm Caldwell <malcolm.caldwell@ntu.edu.au>
 * I hard coded room "loopback" as a loopback room regardless of
 * the --audio-loopback parameter.
 *
 * Revision 1.42  2001/11/27 12:33:56  rogerh
 * Fix bug in statictics which caused a core dump
 *
 * Revision 1.41  2001/11/22 13:06:38  rogerh
 * Add --audio-loopback. Users of OpenMCU can hear their own voice echoed back
 * which is very handy for testing purposes.
 *
 * Revision 1.40  2001/08/27 03:47:39  robertj
 * Changed if statement with constant expression to be #ifdef.
 *
 * Revision 1.39  2001/08/24 13:48:01  rogerh
 * Delete the listener thread if StartListener() fails.
 *
 * Revision 1.38  2001/08/23 07:54:09  rogerh
 * Add code which allows a user to hear their own audio echo back from the MCU
 * This needs to be enabled at compile time with the HEAR_OWN_AUDIO define.
 *
 * Revision 1.37  2001/07/23 03:55:13  rogerh
 * Seperate out codec names for audio and video
 *
 * Revision 1.36  2001/07/23 03:28:03  rogerh
 * Display the codec name in the statistics page
 *
 * Revision 1.35  2001/07/17 11:49:38  rogerh
 * Change title, Mcu -> MCU
 *
 * Revision 1.34  2001/07/17 10:57:45  rogerh
 * Add LPC-10 codec support, from Santiago Garcia Mantinan <openh323@manty.net>
 *
 * Revision 1.33  2001/07/12 06:18:27  rogerh
 * Add G711 A-Law codec
 *
 * Revision 1.32  2001/06/28 06:08:17  rogerh
 * Fix compilation warning
 *
 * Revision 1.31  2001/06/13 06:01:35  rogerh
 * Add some comments
 *
 * Revision 1.30  2001/06/13 05:47:19  rogerh
 * Fix the Make Call menu option which was broken when multiple rooms were
 * added. Calls made from OpenMCU are added to the default room name.
 * Add the Microsoft GSM codec to the codecs list.
 *
 * Revision 1.29  2001/05/31 17:01:52  rogerh
 * Fixes from Dan Johansson <djn98006@student.mdh.se> to make video work.
 *  Go back to using 'closed' for the Video Classes. This is needed as
 *  the the video classes come from PVideoChannel which does not use os_handle
 *  in its IsOpen() method. Instead, we must define our own IsOpen() method.
 *  Also, back out the size of the image change.
 *  Finally, add a new feature. For the first 4 connections, video from an
 *  endpoint is displayed immediatly rather than waiting until that ep sends
 *  some audio. (handy for endpoints with video only and with no talking)
 *
 * Revision 1.28  2001/05/31 14:29:29  rogerh
 * Add --disable-menu to OpenMCU
 *
 * Revision 1.27  2001/05/17 22:07:51  dereks
 * fixed calculation of the number of bytes in the image.
 *
 * Revision 1.26  2001/05/08 13:43:11  rogerh
 * Connections without a room name will now join the default room (room101)
 * Handy for NetMeeting users who cannot specify the room to OpenMCU.
 * Add --defaultroom to change the default room name. Add --no-defaultroom to
 * prevent use of the default room and to reject calls without a room name.
 *
 * Revision 1.25  2001/05/08 11:30:39  rogerh
 * clean up display
 *
 * Revision 1.24  2001/05/08 10:47:07  rogerh
 * Handle the removal of members correctly.
 *
 * Revision 1.23  2001/05/04 08:07:49  rogerh
 * Improve stats when there are no connections and fix nested mutex bug
 *
 * Revision 1.22  2001/03/20 23:34:33  robertj
 * Added creating of room member lists on first use of room.
 * Used the new PTrace::Initialise function for starting trace code.
 *
 * Revision 1.21  2001/03/18 07:40:45  robertj
 * Fixed MSVC compatibility.
 *
 * Revision 1.20  2001/03/18 06:50:20  robertj
 * More changes for multiple conferences from Patrick Koorevaar.
 *
 * Revision 1.19  2001/03/06 04:32:50  robertj
 * Fixed another error in previous update, variable declared twice.
 *
 * Revision 1.18  2001/03/05 22:36:22  robertj
 * Changes for logging and multiple conferences from Patrick Koorevaar.
 *
 * Revision 1.17  2001/02/19 05:04:18  robertj
 * Added more fixes on shutdown, thanks Paul E. Zaremba.
 *
 * Revision 1.16  2001/02/09 06:09:42  robertj
 * Added fix for crashing on exit problem, thanks Dhomin.
 *
 * Revision 1.15  2001/02/08 07:06:37  robertj
 * Added 'm' command to make call, thanks Paul Zaremba.
 * Added ability to send CIF size images, thanks again Paul Zaremba.
 *
 * Revision 1.14  2001/01/04 01:35:43  dereks
 * Add user interface thread to openmcu.
 * tidy up the exiting process, but it is still in need of work.
 *
 * Revision 1.13  2001/01/03 03:59:26  dereks
 * Adjusted algorithm for selecting which corners contain which video stream.
 * Add gsmframes and g711frames option. Add documentation describing data flows.
 *
 * Revision 1.12  2000/12/22 08:28:23  dereks
 * Optimise video handling, and reduce load on mcu computer
 * Include noise detection routine, to determine which images are displayed when > 4 connections.
 *
 * Revision 1.11  2000/12/19 22:41:44  dereks
 * Add video conferencing - limited to 5 connected nodes.
 * Use the video channel facility now in openh323 and pwlib modules
 * Add simple interface to handle commands entered at the keyboard.
 *
 * Revision 1.10  2000/11/08 04:27:33  robertj
 * Fixed MSVC warnings.
 *
 * Revision 1.9  2000/11/03 06:38:07  craigs
 * Added conditional for windows until we decide what to do
 *
 * Revision 1.8  2000/11/02 03:33:41  craigs
 * Changed to provide some sort of software timeing loop
 *
 * Revision 1.7  2000/06/20 02:38:32  robertj
 * Changed H323TransportAddress to default to IP.
 *
 * Revision 1.6  2000/05/25 13:26:18  robertj
 * Fixed incorrect "save" parameter specification.
 *
 * Revision 1.5  2000/05/25 12:06:20  robertj
 * Added PConfigArgs class so can save program arguments to config files.
 *
 * Revision 1.4  2000/05/11 11:47:11  robertj
 * Fixed alpha linux GNU compiler problems.
 *
 * Revision 1.3  2000/05/11 09:54:02  robertj
 * Win32 compilation
 *
 * Revision 1.2  2000/05/10 08:11:57  craigs
 * Fixed copyrights and names
 *
 * Revision 1.1  2000/05/10 05:54:06  craigs
 * Initial version
 *
 */

#include <ptlib.h>
#include <ptlib/pipechan.h>

#include "version.h"
#include "mscodecs.h"
#include "lpc10codec.h"
#include "speexcodec.h"
#include "ilbccodec.h"
#ifndef NO_MCU_VIDEO
#include "h261codec.h"
#include "h263codec.h"
#include "ffh263codec.h"
#endif

#include "main.h"

static PString DEFAULT_CALL_LOG = "c:\\mcu_log.txt"; 
static PFilePath logFilename = DEFAULT_CALL_LOG;

PCREATE_PROCESS(OpenMcu);

#define new PNEW

///////////////////////////////////////////////////////////////

OpenMcu::OpenMcu()
  : PProcess("OpenH323 Project", "OpenMCU",
             MAJOR_VERSION, MINOR_VERSION, BUILD_TYPE, BUILD_NUMBER)
{
}


OpenMcu::~OpenMcu()
{
}


void OpenMcu::Main()
{
  cout << GetName()
       << " Version " << GetVersion(TRUE)
       << " by " << GetManufacturer()
       << " on " << GetOSClass() << ' ' << GetOSName()
       << " (" << GetOSVersion() << '-' << GetOSHardware() << ")\n";
  //PArgList & args = GetArguments();
  PConfigArgs args(GetArguments());

  args.Parse(
             "g-gatekeeper:"         "n-no-gatekeeper."
             "-require-gatekeeper."  "-no-require-gatekeeper."
             "q-translate:"
             "h-help."
             "i-interface:"          "-no-interface."
             "-save."
             "j-jitter:"
             "-listenport:"
             "-gsmframes:"           "-no-gsmframes."
             "-g711frames:"          "-no-g711frames."
             "-ilbc15frames:"        "-no-ilbc15frames."
             "-ilbc13frames:"        "-no-ilbc13frames."
#if PTRACING
             "t-trace."
             "o-output:"
#endif
             "-defaultroom:"         "-no-defaultroom."
             "u-username:"           "-no-username."
             "p-password:"
#ifndef NO_MCU_VIDEO
             "v-video."              "-no-video."
             "-videotxfps:"          "-no-videotxfps."
             "-videobitrate:"        "-no-videobitrate."
             "-videofill:"           "-no-videofill."
             "-videolarge."
             "-videotxquality:"      "-no-videotxquality."
             "-single-stream."
             "-disable-menu."
             "-audio-loopback:"
#endif
          , FALSE);

#if PTRACING
  PTrace::Initialise(args.GetOptionCount('t'),
                     args.HasOption('o') ? (const char *)args.GetOptionString('o') : NULL);
#endif

  if (args.HasOption('h')) {
    cout << "Usage : " << GetName() << " [options]\n"
            "Options:\n"
            "  -u --username str   : Set the local endpoint name to str\n"
            "  -g --gatekeeper host: Specify gatekeeper host.\n"
            "  -n --no-gatekeeper  : Disable gatekeeper discovery.\n"
            "  -p --password str   : Set the gatekeeper password to str\n"
            "  --require-gatekeeper: Exit if gatekeeper discovery fails.\n"
            "  -q --translate ip   : Set external IP address to ip if masQueraded \n"
            "  -i --interface ip   : Bind to a specific interface\n"
            "  -j --jitter [min-]max : Set minimum (optional) and maximum jitter buffer (in milliseconds).\n"
            "  --listenport n      : Port to listen on for incoming connections(default 1720)\n"
            "  --g711frames count  : Set the number G.711 frames in capabilities (default 30)\n"
            "  --gsmframes count   : Set the number GSM frames in capabilities (default 4)\n"
            "  --ilbc15frames count: Set the number iLBC15k2 frames in capabilities (default 3)\n"
            "  --ilbc13frames count: Set the number iLBC13k3 frames in capabilities (default 4)\n"
#if PTRACING
            "  -t --trace          : Enable trace, use multiple times for more detail\n"
            "  -o --output         : File for trace output, default is stderr\n"
#endif
            "  --save              : Save arguments in configuration file\n"
#ifndef NO_MCU_VIDEO
            "  -v --video          : Enable H261 video handling\n"
      	    "  --videotxfps n      : Maximum number of transmitted video frames per sec 1<10(def)<30\n\n"
            "  --videobitrate n    : Enable bitrate control.   16 < n < 2048 kbit/s (net bw)\n"
            "  --videofill n       : Select number of updated background blocks per frame 1<=n<=99 (2 def)\n"
            "  --videolarge        : Set the video size from normal (176x144) to large (352x288).\n"
            "  --videotxquality n  : Select sent video quality,(def 9). 1(good)<=n<=31\n"
            "  --single-stream     : Disable video and audio multiplexing\n"
 #endif
            "  --defaultroom name  : Connections without a room name will join this room\n"
            "                         (Default room is room101)\n"
            "  --no-defaultroom    : Reject connections with no room specified\n"
            "  --disable-menu      : Disable the command line menu\n"
            "  --audio-loopback name : Let a user hear their own voice echo'ed back in this room\n"
            "  -h --help           : Display this help message\n";
    return;
  }

  args.Save("save");

  MyH323EndPoint endpoint;

  PString userName = "OpenH323 MCU v" + GetVersion();
  if (args.HasOption('u'))
    userName = args.GetOptionString('u');
  endpoint.SetLocalUserName(userName);

  if (args.HasOption('p'))
    endpoint.SetGatekeeperPassword(args.GetOptionString('p'));
    
  int g711Frames = 30;
  if (args.HasOption("g711frames")) {
    g711Frames = args.GetOptionString("g711frames").AsInteger();
    if (g711Frames <= 10 || g711Frames > 240) {
      cerr << "error: G.711 frame size must be in range 10 to 240" << endl;
      g711Frames = 30;
    }
  }
  
  int gsmFrames = 4;
  if (args.HasOption("gsmframes")) {
    gsmFrames = args.GetOptionString("gsmframes").AsInteger();
    if (gsmFrames < 1 || gsmFrames > 7) {
      cerr << "error: GSM frame size must be in range 1 to 7" << endl;
      gsmFrames = 4;
    }
  }
  
  int ilbc15Frames = 3;
  if (args.HasOption("ilbc15frames")) {
    ilbc15Frames = args.GetOptionString("ilbc15frames").AsInteger();
    if (ilbc15Frames < 1 || ilbc15Frames > 6) {
      cerr << "error: iLBC-15k2 frame size must be in range 1 to 6" << endl;
      ilbc15Frames = 3;
    }
  }
  
  int ilbc13Frames = 4;
  if (args.HasOption("ilbc13frames")) {
    ilbc13Frames = args.GetOptionString("ilbc13frames").AsInteger();
    if (ilbc13Frames < 1 || ilbc13Frames > 6) {
      cerr << "error: iLBC-13k3 frame size must be in range 1 to 6" << endl;
      ilbc13Frames = 4;
    }
  }
  
  H323Capability * gsm    = new H323_GSM0610Capability;
  //H323Capability * msgsm  = new MicrosoftGSMAudioCapability;
  H323Capability * ulaw64 = new H323_G711Capability(H323_G711Capability::muLaw, H323_G711Capability::At64k);
  H323Capability * alaw64 = new H323_G711Capability(H323_G711Capability::ALaw, H323_G711Capability::At64k);
  /*H323Capability * lpc10  = */ new H323_LPC10Capability(endpoint);
  H323Capability * speex3 = new SpeexNarrow3AudioCapability(); // 8k
  H323Capability * speex5 = new SpeexNarrow5AudioCapability(); // 15k
  H323Capability * ilbc15 = new H323_iLBC_Capability(endpoint,H323_iLBC_Capability::e_15k2);
  H323Capability * ilbc13 = new H323_iLBC_Capability(endpoint,H323_iLBC_Capability::e_13k3);

  endpoint.SetCapability(0, 0, ilbc15);
  ilbc15->SetTxFramesInPacket(ilbc15Frames);
  endpoint.SetCapability(0, 0, ilbc13);
  ilbc13->SetTxFramesInPacket(ilbc13Frames);
 
  endpoint.SetCapability(0, 0, speex3);
  endpoint.SetCapability(0, 0, speex5);

  endpoint.SetCapability(0, 0, gsm);
  gsm->SetTxFramesInPacket(gsmFrames);

  //endpoint.SetCapability(0, 0, msgsm);
  //msgsm->SetTxFramesInPacket(gsmFrames);

  endpoint.SetCapability(0, 0, ulaw64);
  ulaw64->SetTxFramesInPacket(g711Frames);

  endpoint.SetCapability(0, 0, alaw64);
  alaw64->SetTxFramesInPacket(g711Frames);

  //endpoint.SetCapability(0, 0, lpc10);

  if (args.HasOption('j')) {
    unsigned minJitter;
    unsigned maxJitter;
    PStringArray delays = args.GetOptionString('j').Tokenise(",-");
    if (delays.GetSize() == 1) {
      minJitter = delays[0].AsUnsigned();
      maxJitter = delays[1].AsUnsigned();
    }
    else {
      maxJitter = delays[0].AsUnsigned();
      minJitter = PMIN(endpoint.GetMinAudioJitterDelay(), maxJitter);
    }
    if (minJitter >= 20 && minJitter <= maxJitter && maxJitter <= 1000)
      endpoint.SetAudioJitterDelay(minJitter, maxJitter);
    else {
      cerr << "Jitter should be between 20 and 1000 milliseconds." << endl;
      return;
    }
  }

  endpoint.Initialise(args);

  // start the H.323 listener
  H323ListenerTCP * listener;

  WORD listenPort = H323EndPoint::DefaultTcpPort;
  if (args.HasOption("listenport"))
    listenPort = (WORD)args.GetOptionString("listenport").AsInteger();

  if (args.GetOptionString('i').IsEmpty()) {
    PIPSocket::Address interfaceAddress(INADDR_ANY);
    listener  = new H323ListenerTCP(endpoint, interfaceAddress, listenPort);
  } else {
    PIPSocket::Address interfaceAddress(args.GetOptionString('i'));
    listener  = new H323ListenerTCP(endpoint, interfaceAddress, listenPort);
  }
  if (!endpoint.StartListener(listener)) {
    cout <<  "Could not open H.323 listener port on "
         << listener->GetListenerPort() << endl;
    delete listener;
    return;
  }

  cout << "Listening on port " << listener->GetListenerPort() << endl;

  if (args.GetOptionString('q').IsEmpty()) {
    endpoint.behind_masq = FALSE;
  } else {
    endpoint.masqAddressPtr = new PIPSocket::Address(args.GetOptionString('q'));
    endpoint.behind_masq = TRUE;
    cout << "Masquerading as address " << *(endpoint.masqAddressPtr) << endl;
  }

#ifndef NO_MCU_VIDEO
  // If videoLarge was specified
  endpoint.videoLarge = args.HasOption("videolarge");

  // get singleStream flag
  endpoint.singleStream = args.HasOption("single-stream");
         
  if (args.HasOption('v')) {
    //Add capability to allow the reception of video.

    // this is the only way I know of at present to transmit
    // CIF size video - pez
    // Do not allow CIF video if size is medium
    if (endpoint.videoLarge)
      endpoint.SetCapability(0, 1, new H323_H261Capability(0, 4, FALSE, FALSE, 6217)); // CIF
      endpoint.SetCapability(0, 1, new H323_H261Capability(2, 0, FALSE, FALSE, 6217)); // QCIF
  } 

  endpoint.EnableVideoReception(args.HasOption('v'));

  int videoTxQual = 10;
  if (args.HasOption("videotxquality")) 
      videoTxQual = args.GetOptionString("videotxquality").AsInteger();
  endpoint.videoTxQuality = PMAX(1, PMIN(31, videoTxQual));

  int videoF = 2;
  if (args.HasOption("videofill")) 
    videoF = args.GetOptionString("videofill").AsInteger();
  endpoint.videoFill = PMAX(1, PMIN(99, videoF));

  int videoFPS = 10;
  if (args.HasOption("videotxfps")) 
    videoFPS = args.GetOptionString("videotxfps").AsInteger();
  endpoint.videoFramesPS = PMAX(1,PMIN(30,videoFPS));

  int videoBitRate = 0; //disable setting videoBitRate.
  if (args.HasOption("videobitrate")) {
    videoBitRate = args.GetOptionString("videobitrate").AsInteger();
    videoBitRate = 1024 * PMAX(16, PMIN(2048, videoBitRate));
  }
  endpoint.videoBitRate = videoBitRate;

  if (args.HasOption('v')) {
#if H323_AVCODEC
    PTRACE(3, "Video bitrate for ffmpeg h263 is " << videoBitRate);
    if (endpoint.videoLarge) 
      endpoint.SetCapability(0, 1, new H323_FFH263Capability(0, 0, 1, 0, 0, videoBitRate, videoFPS));
    endpoint.SetCapability(0, 1, new H323_FFH263Capability(0, 1, 0, 0, 0, videoBitRate, videoFPS));
#endif
#if H323_VICH263
    PTRACE(3, "Video bitrate for vic h263 is " << videoBitRate);
    if (endpoint.videoLarge) 
      endpoint.SetCapability(0, 1, new H323_H263Capability(0, 0, 1, 0, 0, videoBitRate, videoFPS));
    endpoint.SetCapability(0, 1, new H323_H263Capability(0, 1, 0, 0, 0, videoBitRate, videoFPS));
#endif
  }
#endif   /*End of test on NO_MCU_VIDEO */


  // add user input capabilities
  endpoint.SetCapability(0, P_MAX_INDEX, new H323_UserInputCapability(H323_UserInputCapability::BasicString));

  if (args.HasOption('g')) {
    PString gkName = args.GetOptionString('g');
    if (endpoint.SetGatekeeper(gkName, new H323TransportUDP(endpoint)))
      cout << "Gatekeeper set: " << *endpoint.GetGatekeeper() << endl;
    else {
      cout << "Error registering with gatekeeper at \"" << gkName << '"' << endl;
      return;
    }
  }
  else if (!args.HasOption('n')) {
    cout << "Searching for gatekeeper..." << flush;
    if (endpoint.DiscoverGatekeeper(new H323TransportUDP(endpoint)))
      cout << "\nGatekeeper found: " << *endpoint.GetGatekeeper() << endl;
    else {
      cout << "\nNo gatekeeper found." << endl;
      if (args.HasOption("require-gatekeeper"))
        return;
    }
  }  

  cout <<  "Codecs (in preference order):\n" << setprecision(2) << endpoint.GetCapabilities() << endl << endl;

  endpoint.ListenForIncomingCalls();
  endpoint.AwaitTermination();
  cout << "GoodBye. " << endl;

  PThread::Current()->Sleep(2000);
  
#if 0 //PTRACING
  if (output != NULL) 
    delete output;
#endif
}

///////////////////////////////////////////////////////////////

MyH323EndPoint::MyH323EndPoint()
{
	terminalType = e_MCUWithAudioMP;
#ifndef NO_MCU_VIDEO
	terminalType = e_MCUWithAVMP;
#endif
  conferenceListener = NULL;
}

void MyH323EndPoint::Initialise(PArgList & args)
{
  // get default "room" (conference) name
  defaultRoomName = "room101";
  if (args.HasOption("defaultroom"))
    defaultRoomName = args.GetOptionString("defaultroom");
  if (args.HasOption("no-defaultroom"))
    defaultRoomName = "";

  // see if menu is required
  hasMenu = !args.HasOption("disable-menu");

  // see if audio loopback is required
  audioLoopbackRoom = args.GetOptionString("audio-loopback");
}

H323Connection * MyH323EndPoint::CreateConnection(unsigned callReference)
{
  return new OpenMCUH323Connection(*this, callReference);
}


void MyH323EndPoint::TranslateTCPAddress(PIPSocket::Address &localAddr, const PIPSocket::Address &remoteAddr)
{

  if (this->behind_masq && 
      (remoteAddr.Byte1() != 192)) {
    localAddr = *(this->masqAddressPtr);
  }
  return;
}

void MyH323EndPoint::ListenForIncomingCalls()
{
  cout << "Waiting for incoming calls for \"" << GetLocalUserName() << '"' << endl;
}


void MyH323EndPoint::AwaitTermination()
{
  PThread * userInterfaceThread = NULL;
  if (hasMenu)
    userInterfaceThread = new UserInterfaceThread(*this);

  // poll the handset every 100ms looking for state changes
  while (!exitFlag.Wait(100)) {

    // lock the user interface state whilst we change it
  }

  if (userInterfaceThread != NULL) {
    userInterfaceThread->Terminate();
    userInterfaceThread->WaitForTermination();
    delete userInterfaceThread;

    cout << "User Interface thread has terminated successfully."<<endl;  
  }

}

void MyH323EndPoint::HandleUserInterface()
{
  PConsoleChannel console(PConsoleChannel::StandardInput);
 
  PTRACE(2, "OpenMCU\tUser interface thread started.");

  PStringStream help;
  help << "Select:\n"
          " ? : Print this help message\n"
#ifndef NO_MCU_VIDEO
          " v : report  connections are at which video corner\n"
#endif
          " l : listen to a conference (no arg gives current conf)\n"
          " u : unlisten to current conference\n"
					" m : make a call\n"
          " x : Exit immediately\n"
	        " s : Report statistics\n"
          " z : Put message in log file (for debugging)\n";
  
  PString user;
  PString str; 

  for (;;) {
    // display the prompt
    cout << "Command ? " << flush;

    
    //terminate the menu loop if console finished
    char ch = (char)console.peek();
    if (console.eof()) {
        cerr << "\nConsole gone - menu disabled" << endl;
        cerr << "Finishing program"<<endl;        
        ch = 'x';
    } else
      console >> ch;

    switch (tolower(ch)) {
      case '?' :
        cout << help << endl;
        break;

#ifndef NO_MCU_VIDEO
      case 'v' : 
        cout << " video information now available with option s" << endl;
        break;
#endif    
      case 'z' : 
        console >> str;
        PTRACE(0,"Log message=" << str);
        break;

      case 'l' :
      	console >> str;
        str = str.Trim();
        if (str.IsEmpty()) {
          if (conferenceListener != NULL)
            cout << "Currently listening to conference " << conferenceListener->GetConference()->GetNumber() << endl;
          else
            cout << "Not listening to any conference" << endl;
        }
        else {
          PWaitAndSignal m(roomListMutex);
          ConferenceList::iterator r;
          for (r = roomList.begin(); (conferenceListener == NULL) && (r != roomList.end()); ++r) {
            if (r->second->GetNumber() == str) {
              conferenceListener = new ConferenceListener(*this);
              r->second->AddMember(conferenceListener);
              cout << "listening to conference " << str << endl;
              break;
            }
          }
          if (r == roomList.end())
            cout << "Could not find conference " << str << endl;
        }
        break;

      case 'u':
        if (conferenceListener == NULL) {
          cout << "Not listening to any conference" << endl;
        } else {
          cout << "stopping listen to conference " << conferenceListener->GetConference()->GetNumber() << endl;
          conferenceListener->Unlisten();
          delete conferenceListener;
          conferenceListener = NULL;
        }
        break;

      case 'q' :
      case 'x' :
        cout << "Exiting." << endl;
			  ClearAllCalls(H323Connection::EndedByLocalUser, FALSE);
        cout << "All calls cleared"<<endl;      
        exitFlag.Signal();
        console.ignore(INT_MAX, '\n');
        return;

      case 'm' :
      	console >> user;
        if (!MakeCall(user, str))
          cerr << "Error making call to \"" << user << '"' << endl;
        break;

      case 's' :
        {
          PWaitAndSignal m(roomListMutex);
          PINDEX roomSize = roomList.size();
          if (roomSize == 0)
            cout << "No statistics available" << endl;
          else {
            PINDEX idx = 1;
            ConferenceList::iterator r;
            for (r = roomList.begin(); r != roomList.end(); ++r) {
              Conference & conference = *(r->second);
              PWaitAndSignal m(conference.GetMutex());
              Conference::MemberList & memberList = conference.GetMemberList();
              cout << "Statistics for " << memberList.size() << " members of conference "
                   << idx++ << " of "<< roomSize << " with ID " 
                   << conference.GetNumber() << endl;

              Conference::MemberList::const_iterator r;
              for (r = memberList.begin(); r != memberList.end(); r++) {
                PString cls = r->second->GetClass();
                if (r->second->IsDescendant(H323Connection_ConferenceMember::Class())) {
                  OpenMCUH323Connection & conn = *((H323Connection_ConferenceMember *)(r->second))->GetConnection();
                  PTime now;
                  PTime callStart = conn.GetConnectionStartTime();
                  cout << "Connected to : " << conn.GetRemotePartyName() << " for " 
                       << setw(5) << setprecision(0) << (now - callStart) << " mins " <<endl;

                  RTP_Session * session = conn.GetSession(RTP_Session::DefaultAudioSessionID);
                  if (session == NULL) {
		                cout << "             : "
		                     << "Sending audio with "
		                     << conn.GetAudioTransmitCodecName()
		                     << endl; 
		                cout << "             : "
		                     << "Receiving audio with "
		                     << conn.GetAudioReceiveCodecName()
		                     << endl;
                  } else {
                    cout << "             : "
                         << session->GetPacketsSent() << '/'
                         << session->GetOctetsSent() 
                         << " audio packets/bytes sent ("
		                     << conn.GetAudioTransmitCodecName()
		                     << ")" 
                         << endl;
                    cout << "             : "
                         << session->GetPacketsReceived() << '/'
                         << session->GetOctetsReceived()
                         << " audio packets/bytes received ("
		                     << conn.GetAudioReceiveCodecName()
		                     << ")" 
                         << endl;
		              }
  #ifndef NO_MCU_VIDEO
		              session = conn->GetSession(RTP_Session::DefaultVideoSessionID);
		              if (session == NULL) {
		                cout << "             : "
		                     << "Sending video with "
		                     << conn->GetVideoTransmitCodecName()
		                     << endl; 
		                cout << "             : "
		                     << "Receiving video with "
		                     << conn->GetVideoReceiveCodecName()
		                     << endl;
		              } else {
                    cout << "             : "
                         << session->GetPacketsSent() << '/'
                         << session->GetOctetsSent() 
                         << " video packets/bytes sent ("
		                     << conn->GetVideoTransmitCodecName()
		                     << ")"
                         << endl;
                    cout << "             : "
                         << session->GetPacketsReceived() << '/'
                         << session->GetOctetsReceived()
                         << " video packets/bytes received ("
		                     << conn->GetVideoReceiveCodecName()
		                     << ")" 
                         << endl;
		              }
  #endif
                }
    #ifndef NO_MCU_VIDEO
                PStringList & videoPosnList = videoPosnDict[roomID];
                for(PINDEX j = 0; j < videoPosnList.GetSize(); j++) {
                  str = videoPosnList[j];
                  str.Delete(0,3);
                  cout << "   video for position " << j << " is \""
                       << str.Left(str.Find(':')) << '"' << endl;
                }
    #endif
              }
            }
          }
        }
        break;

      default :
        cout << "The input character \""<<ch<<"\" is not defined"<<endl;
		}
  }
}

PString MyH323EndPoint::GetNewRoomNumber()
{
  static PMutex mutex;
  static unsigned number = 100;

  PWaitAndSignal m(mutex);
  return PString(PString::Unsigned, number++);
}

Conference * MyH323EndPoint::ConferenceRequest(H323Connection & connection, 
                                          const H323SignalPDU & setupPDU,
                           H323Connection::AnswerCallResponse & response)
{
  const H225_Setup_UUIE & setup = setupPDU.m_h323_uu_pdu.m_h323_message_body;

  // get the conference ID from the incoming call
  OpalGloballyUniqueID conferenceID = setup.m_conferenceID;

  // we accept creates and joins. Invites come later :)
  if (setup.m_conferenceGoal.GetTag() != H225_Setup_UUIE_conferenceGoal::e_create &&
      setup.m_conferenceGoal.GetTag() != H225_Setup_UUIE_conferenceGoal::e_join) {
    PTRACE(3, "Conference\tUnknown conference goal " << setup.m_conferenceGoal.GetTag());
    response = H323Connection::AnswerCallDenied;
    return NULL;
  }

  roomListMutex.Wait();

  // look for the conference ID in the current conference list 
  // just in case the endpoint knows the conference it wants
  Conference * conference = NULL;
  ConferenceList::iterator r = roomList.find(conferenceID);
  if (r != roomList.end())
    conference = r->second;

  // if the endpoint was specifically asking to join a conference, and that 
  // conference is unknown, then return ReleaseComplete with the appropriate reason
  if (conference == NULL && (setup.m_conferenceGoal.GetTag() == H225_Setup_UUIE_conferenceGoal::e_join)) {
    roomListMutex.Signal();
    response = H323Connection::AnswerCallDeniedByInvalidCID;
    PTRACE(3, "Conference\tRequest to join nonexistent conference " << conferenceID);
    return NULL;
  }

  // nominally, H.323 version 2 terminals can do conference lists, 
  BOOL canDoConferenceList = connection.GetSignallingVersion() >= 2;

  PString appInfo = connection.GetRemoteApplication();

  // we know NetMeeting cannot do conference lists
  if (appInfo.Find("NetMeeting") != P_MAX_INDEX)
    canDoConferenceList = FALSE;

  // we know OpenH323 apps prior to 1.14 cannot handle conference 
  else if (appInfo.Find("OpenH323") != P_MAX_INDEX) {
    PINDEX pos = appInfo.Find("OpenH323 ");
    if (pos != P_MAX_INDEX) {
      PStringArray version = appInfo.Mid(pos+9).Tokenise('.');
      if (!(version.GetSize() == 3 &&
            version[0].AsUnsigned() >= 1 &&
            version[1].AsUnsigned() >= 14 &&
            version[2].AsUnsigned() >= 0))
        canDoConferenceList = FALSE;
    }
  }

  // get the name of the room to create
  PString roomToCreate = defaultRoomName;

  // if the endpoint cannot do conference lists, 
  // then see if the called party number matches a room number
  if (!canDoConferenceList && (conference == NULL) && (setup.m_destinationAddress.GetSize() > 0)) {
    PINDEX i;
    for (i = 0; (conference == NULL) && (i < setup.m_destinationAddress.GetSize()); i++) {
      PString roomNumber = H323GetAliasAddressString(setup.m_destinationAddress[i++]);
      for (r = roomList.begin(); (conference == NULL) && (r != roomList.end()); ++r) {
        if (r->second->GetNumber() == roomNumber) {
          conference = r->second;
          break;
        }
      }
    }

    // if we did not find a matching room, then we need to create a new one
    if (conference == NULL)
      roomToCreate = H323GetAliasAddressString(setup.m_destinationAddress[0]);
  }

  // if no conference, and no default room, then send the terminal a conference list if it can understand it
  if (conference == NULL && roomToCreate.IsEmpty() && canDoConferenceList) {

    H323SignalPDU facilityPDU;
    H225_Facility_UUIE & pdu = *facilityPDU.BuildFacility(connection, FALSE);

    //facilityPDU.RemoveOptionalField(H225_Facility_UUIE::e_h245Tunneling);

    pdu.RemoveOptionalField(H225_Facility_UUIE::e_callIdentifier);
    pdu.RemoveOptionalField(H225_Facility_UUIE::e_multipleCalls);
    pdu.RemoveOptionalField(H225_Facility_UUIE::e_maintainConnection);

    pdu.m_reason = H225_FacilityReason::e_conferenceListChoice;
    pdu.IncludeOptionalField(H225_Facility_UUIE::e_conferences);

    H225_ArrayOf_ConferenceList & conferenceList = pdu.m_conferences;
    conferenceList.SetSize(0);

    PINDEX i = 0;
    for (r = roomList.begin(); r != roomList.end(); ++r) {
      Conference & conf = *(r->second);
      if (conf.IsVisible()) {
        conferenceList.SetSize(conferenceList.GetSize()+1);
        H225_ConferenceList & conference = conferenceList[i++];

        conference.IncludeOptionalField(H225_ConferenceList::e_conferenceID);
        conference.m_conferenceID = conf.GetID();

        conference.IncludeOptionalField(H225_ConferenceList::e_conferenceAlias);
        H323SetAliasAddress(conf.GetName(), conference.m_conferenceAlias, -1);
      }
    }

    roomListMutex.Signal();

    PTRACE(3, "Conference\tSending conference list");

    // if the PDU write fails, then close the call
    if (!connection.WriteSignalPDU(facilityPDU))
      response = H323Connection::AnswerCallDenied;
    else 
      response = H323Connection::AnswerCallDeferred;

    return NULL;
  }

  // if there is a room to create, then join this call to that conference
  if (!roomToCreate.IsEmpty()) {
    PTRACE(3, "Conference\tCreating/joining default conference room " << roomToCreate);
    for (r = roomList.begin(); r != roomList.end(); ++r) {
      Conference & conf = *(r->second);
      if (conf.GetNumber() == roomToCreate) {
        conference = &conf;
        break;
      }
    }
    if (conference == NULL) {
      conference = new Conference(conferenceID, roomToCreate, "");
      roomList.insert(std::pair<OpalGloballyUniqueID, Conference *>(conferenceID, conference));
    }
  }

  // else create a new conference with a new room number
  else {
    PTRACE(3, "Conference\tCreating new conference");
    conference = new Conference(conferenceID, GetNewRoomNumber(), "");
    roomList.insert(std::pair<OpalGloballyUniqueID, Conference *>(conferenceID, conference));
  }

  roomListMutex.Signal();

  return conference;
}

void MyH323EndPoint::RemoveConference(const OpalGloballyUniqueID & confId)
{
  PWaitAndSignal m(roomListMutex);
  ConferenceList::iterator r = roomList.find(confId);
  if (r != roomList.end())  {
    Conference * conf = r->second;
    roomList.erase(confId);
    delete conf;
  }
}

///////////////////////////////////////////////////////////////

OpenMCUH323Connection::OpenMCUH323Connection(MyH323EndPoint & _ep, unsigned callReference)
  : H323Connection(_ep, callReference), ep(_ep), connected(FALSE)
{
  incomingAudio    = NULL;
  outgoingAudio    = NULL;
  conference       = NULL;
  conferenceMember = NULL;
  sendingAllowed   = TRUE;

  audioReceiveCodecName = audioTransmitCodecName = "none";

#ifndef NO_MCU_VIDEO
  incomingVideo = NULL;
  outgoingVideo = NULL;
  videoReceiveCodecName = videoTransmitCodecName = "none";
#endif

  connected = FALSE;

  cout << "Opening connection" << endl;
}

OpenMCUH323Connection::~OpenMCUH323Connection()
{
  cout << "Closing connection" << endl;
}

void OpenMCUH323Connection::CleanUpOnCallEnd()
{
  if (conference != NULL && conferenceMember != NULL) {
    LogCall();
    conferenceMember->WaitForClose();
    ConferenceMember * cmToDelete = conferenceMember;
    conferenceMember = NULL;
    if (conference->RemoveMember(cmToDelete))
      ep.RemoveConference(conference->GetID());
    delete cmToDelete;
  }

  H323Connection::CleanUpOnCallEnd();
}

H323Connection::AnswerCallResponse OpenMCUH323Connection::OnAnswerCall(const PString & /*caller*/,
                                                                  const H323SignalPDU & setupPDU,
                                                                  H323SignalPDU & /*connectPDU*/)
{
  H323Connection::AnswerCallResponse response;
  conference = ep.ConferenceRequest(*this, setupPDU, response);

  if (conference == NULL)
    return response;

  connected = TRUE;
  conferenceMember = new H323Connection_ConferenceMember(this);
  conference->AddMember(conferenceMember);

  return AnswerCallNow;
}


BOOL OpenMCUH323Connection::OnSendSignalSetup( H323SignalPDU & callProceedingPDU)
{
  // We are making a connection to a remote EP so add this connection to
  // the list of rooms and members.
  // We will add them to the default room as we have no method of
  // specifying which room our connection should join.

  connected = TRUE;
  conference = NULL;
  PTRACE(3, "Conference\tOutgoing connection");

  return H323Connection::OnSendSignalSetup( callProceedingPDU );
}


BOOL OpenMCUH323Connection::OpenAudioChannel(BOOL isEncoding, unsigned /* bufferSize */, H323AudioCodec & codec)
{
  PStringStream codecName;
  codecName << codec;

  codec.SetSilenceDetectionMode( H323AudioCodec::NoSilenceDetection );

//  if (!codec.IsDescendant(H323_GSM0610Codec::Class()) &&
//      - need to add MS-GSM here along with any other codecs
//      !codec.IsDescendant(H323_muLawCodec::Class())) {
//    cerr << "Unknown codec \"" << codecName << endl;
//    return FALSE;
//  }

  if ((incomingAudio == NULL) && sendingAllowed) {
    incomingAudio = new IncomingAudio(ep, *this);
  }

  if (outgoingAudio == NULL) {
    outgoingAudio = new OutgoingAudio(ep, *this);
  }

  if (isEncoding) {
    audioTransmitCodecName = codecName;
    codec.AttachChannel(outgoingAudio, TRUE);
  } else if (sendingAllowed) {
    audioReceiveCodecName = codecName;
    codec.AttachChannel(incomingAudio, TRUE);
  }

  return TRUE;
}

#ifndef NO_MCU_VIDEO
BOOL OpenMCUH323Connection::OpenVideoChannel(BOOL isEncoding,
                                        H323VideoCodec & codec)
{
    PStringStream codecName;
    codecName << codec;

    PWaitAndSignal mutex(videoMutex);

    if (isEncoding) {
      if (outgoingVideo == NULL) {
        outgoingVideo = new OutgoingVideo(ep, *this, ep.videoFramesPS, ep.videoLarge);
        codec.AttachChannel(outgoingVideo,TRUE);
//       outgoingVideo->SetFrameSize(352>>1,288>>1);
      }

      /*At last. Modularity. The video codec is told the parameters of video compresion/decompression.
        The only thing the video codec knows about the ouside world is how to acquire/render data, which
	      is via the video channel, provided by the OutgoingVideo class.
	      
	      The codec does provide a second interface, through which ethernet packets enter (or leave) */
      videoTransmitCodecName = codecName;

      codec.SetTxQualityLevel(ep.videoTxQuality);
      codec.SetBackgroundFill(ep.videoFill);
      if (ep.videoBitRate != 0)
        codec.SetMaxBitRate(ep.videoBitRate);

    } else {
      if ((incomingVideo == NULL) && sendingAllowed)
        incomingVideo = new IncomingVideo(ep, *this);
        codec.AttachChannel(incomingVideo,TRUE);
        videoReceiveCodecName = codecName;
    }
  
    return TRUE;
}
#endif

void OpenMCUH323Connection::OnUserInputString(const PString & str)
{
  if (conferenceMember != NULL)
    conferenceMember->OnReceivedUserInputIndication(str);
}


BOOL OpenMCUH323Connection::OnIncomingAudio(const void * buffer, PINDEX amount)
{
  if (conferenceMember != NULL)
    conferenceMember->OnIncomingAudio(buffer, amount);

  return TRUE;
}

BOOL OpenMCUH323Connection::OnOutgoingAudio(void * buffer, PINDEX amount)
{
  if (conferenceMember != NULL)
    conferenceMember->OnOutgoingAudio(buffer, amount);

  return TRUE;
}

#ifndef NO_MCU_VIDEO

BOOL OpenMCUH323Connection::OnOutgoingVideo(void * buffer, PINDEX & amount)
{
  peerList.OnOutgoingVideo(buffer, amount);
} 

BOOL OpenMCUH323Connection::OnIncomingVideo(const void * buffer, PINDEX amount)
{
  return conference->OnIncomingVideo(buffer, amount);
}

#endif

///////////////////////////////////////////////////////////////

static void LogMessage(const PString & str)
{
  static PMutex logMutex;
  static PTextFile logFile;

  PTime now;
  PString msg = now.AsString("dd/MM/yyyy") & str;
  logMutex.Wait();

  if (!logFile.IsOpen()) {
    logFile.Open(logFilename, PFile::ReadWrite);
    logFile.SetPosition(0, PFile::End);
  }

  logFile.WriteLine(msg);
  logFile.Close();
  logMutex.Signal();
}

void OpenMCUH323Connection::LogCall(const BOOL accepted)
{
  H323TransportAddress address = GetControlChannel().GetRemoteAddress();
  PIPSocket::Address ip;
  WORD port;
  PStringStream stringStream, timeStream;
  address.GetIpAndPort(ip, port);
  timeStream << GetConnectionStartTime().AsString("hh:mm:ss");
  stringStream << ' ' << "caller-ip:" << ip << ':' << port << ' '
	             << GetRemotePartyName() 
               << " room:" << ((conference != NULL) ? conference->GetNumber() : PString());

  if (accepted) {
    PStringStream connectionDuration;
    connectionDuration << setprecision(0) << setw(5) << (PTime() - GetConnectionStartTime());
    LogMessage(timeStream + stringStream	+ " connection duration:" + connectionDuration);
  }
  else 
    LogMessage(timeStream + " Call denied:" + stringStream);		
}

///////////////////////////////////////////////////////////////

OutgoingAudio::OutgoingAudio(H323EndPoint & _ep, OpenMCUH323Connection & _conn)
  : ep(_ep), conn(_conn)
{
  os_handle = 0;
}

void OutgoingAudio::CreateSilence(void * buffer, PINDEX amount)
{
  memset(buffer, 0, amount);
  lastReadCount = amount;
}

BOOL OutgoingAudio::Read(void * buffer, PINDEX amount)
{
  PWaitAndSignal mutexR(audioChanMutex);
  
  if (!IsOpen())
    return FALSE;

  if (!delay.Delay(amount / 16)) {

    // do the read call here, by calling conn.OnOutgoingAudio():
    BOOL doSilence = !conn.OnOutgoingAudio(buffer, amount);

    if (doSilence)
      CreateSilence(buffer, amount);
  }

  lastReadCount = amount;

  return TRUE;
}

BOOL OutgoingAudio::Close()
{
  if (!IsOpen()) 
    return FALSE;

  PWaitAndSignal mutexC(audioChanMutex);
  os_handle = -1;

  return TRUE;
}

///////////////////////////////////////////////////////////////////////////

IncomingAudio::IncomingAudio(H323EndPoint & _ep, OpenMCUH323Connection & _conn)
  : ep(_ep), conn(_conn)
{
  os_handle = 0;
}

BOOL IncomingAudio::Write(const void * buffer, PINDEX amount)
{
  PWaitAndSignal mutexW(audioChanMutex);
  
  if (!IsOpen())
    return FALSE;

  if (!delay.Delay(amount / 16))
    conn.OnIncomingAudio(buffer, amount);

  return TRUE;
}

BOOL IncomingAudio::Close()
{
  if (!IsOpen())
    return FALSE;

  PWaitAndSignal mutexA(audioChanMutex);
  os_handle = -1;

  return TRUE;
}

///////////////////////////////////////////////////////////////////////////

ConferenceListener::ConferenceListener(MyH323EndPoint & _ep)
#ifdef _WIN32
#pragma warning(push)
#pragma warning(disable:4355)
#endif
  : ConferenceMember(this), ep(_ep)
#ifdef _WIN32
#pragma warning(pop)
#endif
{
  // open the default audio device
  PString deviceName = PSoundChannel::GetDefaultDevice(PSoundChannel::Player);
  soundDevice.Open(deviceName, PSoundChannel::Player, 1, 8000, 16);

  // start a thread to read from the conference and write to default audio device
  if (!soundDevice.IsOpen()) {
    thread = NULL;
  } else {
    running = TRUE;
    thread = PThread::Create(PCREATE_NOTIFIER(Thread), 0, PThread::NoAutoDeleteThread);
  }
}

ConferenceListener::~ConferenceListener()
{
  Unlisten();
}

void ConferenceListener::Unlisten()
{
  if (conference->RemoveMember(this))
    ep.RemoveConference(conference->GetID());

  if ((thread != NULL) && running) {
    running = FALSE;
    thread->WaitForTermination();
    delete thread;
    thread = NULL;
    ep.RemoveConferenceListener();
  }
}

void ConferenceListener::Thread(PThread &, INT)
{
  AudioDelay audioDelay;

  PBYTEArray pcmData(480);
  soundDevice.SetBuffers(480, 3);
  while (running) {
    if (!soundDevice.IsOpen())
      PThread::Sleep(100);
    else if (!lock.Wait())
      PThread::Sleep(30);
    else {

      // First, set the buffer to empty.
      memset(pcmData.GetPointer(), 0, pcmData.GetSize());

      // get number of channels to mix
      ConnectionList::iterator r;
      for (r = connectionList.begin(); r != connectionList.end(); ++r) {
        r->second->ReadAudio(pcmData.GetPointer(), pcmData.GetSize());
        //r->second->ReadAndMixAudio(pcmData.GetPointer(), pcmData.GetSize(), connectionList.size());
      }

      lock.Signal();
      soundDevice.Write(pcmData.GetPointer(), pcmData.GetSize());
      audioDelay.Delay(pcmData.GetSize() / 16);
    }
  }
}

// End of File ///////////////////////////////////////////////////////////////
